/*
 * This module is part of a controllable faulty router implementation
 * 
 * Completed as part of a final year project at University of Wales, Aberystwyth
 * Thomas Salmon - tws9@aber.ac.uk (tom@slashtom.org)
 *
 * (C) 2003 Tom Salmon, this code is GPL
 */


#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/in.h>

#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#define PROTO_ICMP 	1
#define PROTO_IGMP  2
#define PROTO_TCP	6
#define PROTO_UDP	17
#define PROTO_OSPF	89

int queue;
static int freq = -1; /* frequency to drop, default value */
static int drop_this = 0; 
static int reorder_freq = -1;
static int reorder_delay = -1;
static int reorder_this = 0;
static int delay_this = -1;

static int reorder_block = -1; /* sizeof block to be reordered */
static int block_size = 1;

/* the delayed packet */
struct delayed_packet {
	struct sk_buff *delayed_skb;
	struct nf_info *delayed_info;
	struct delayed_packet *next;
};

struct delayed_packet *delayed = NULL;

/* the data passed in from user space program */
struct freq_struct {
	int freq;
	int reorder_freq;
	int reorder_delay;
	int reorder_block;
};

/* function called when packet arrives for this hook */
/* NF_QUEUE sends packets to the queue, also in this module */
static unsigned int faulty(unsigned int hook,
			struct sk_buff **skb,
			const struct net_device *in,
			const struct net_device *out,
			int (*okfn)(struct sk_buff *)){
	if (freq < 0){
		return NF_QUEUE;
	}
	if (skb[0]->h.ipiph->protocol == PROTO_ICMP){
		if (drop_this == freq){
			drop_this = 1;
			return NF_DROP;
		}
		else{
			drop_this++;
		}
	}
	return NF_QUEUE;
}

/* inserts a delayed packet into linked list of delayed packets */
void insert_packet(struct delayed_packet *d, 
			struct sk_buff *skb, struct nf_info *info){
	/* sanity check */
	if (d == NULL){
		printk("ERROR: insert_packet failed sanity check\n");
		return;
	}

	if (d->next == NULL){
		d->next = kmalloc(sizeof(struct delayed_packet), GFP_ATOMIC);
		/*d->next->delayed_skb = kmalloc(sizeof(struct sk_buff), GFP_ATOMIC);
		d->next->delayed_info = kmalloc(sizeof(struct nf_info), GFP_ATOMIC);
		*/
		d->next->delayed_skb = skb_copy(skb, GFP_ATOMIC);
		d->next->delayed_info = info;
		d->next->next = NULL;
		printk("Delaying packet\n");
	}
	else{
		insert_packet(d->next, skb, info);
	}
}

/* free's the memory used be the delayed packet list */
void free_delayed(struct delayed_packet *d){
	if (d->next != NULL){
		free_delayed(d->next);
		kfree(d->next);
	}
	return;
}

ip_fault_queue(struct sk_buff *skb, struct nf_info *info, void *data){
	if (reorder_freq > 0){
		if (skb->h.ipiph->protocol == PROTO_ICMP){
			if (reorder_this == reorder_freq){
				if (delayed == NULL){
					delayed = kmalloc(sizeof(struct delayed_packet), GFP_ATOMIC);
					/*
					delayed->delayed_skb = kmalloc(sizeof(struct sk_buff), GFP_ATOMIC);
					delayed->delayed_info = kmalloc(sizeof(struct nf_info), GFP_ATOMIC);
					*/
					delayed->delayed_skb = skb;
					delayed->delayed_info = info;
					delayed->next = NULL;
				}
				else{
					insert_packet(delayed, skb, info);
				}

				if (reorder_block == block_size && reorder_block > 1){
					/* reset counter */
					reorder_this = 1;
					block_size = 1;
					delay_this = reorder_delay-1;
				}
				else {
					block_size++;
				}
				return 0;
			}
			else{
				reorder_this++;
			}
			if (delay_this == 0){
				struct delayed_packet *d = delayed;

				while(d != NULL){
					/* sanity check */
					if (delayed == NULL ||
						delayed->delayed_skb == NULL || 
						delayed->delayed_info == NULL){
						printk("ERROR: Null in delayed\n");
						nf_reinject(skb, info, NF_ACCEPT);
						return 0;
					}

					printk("Inserting delayed packet\n");
					nf_reinject(d->delayed_skb, d->delayed_info, NF_ACCEPT);
					/*
					kfree(delayed->delayed_skb);
					kfree(delayed->delayed_info);
					*/

					if (d->next == NULL){
						printk("Next node is null\n");
						break;
					}
					d = d->next;
				}
				free_delayed(delayed);
				kfree(delayed);
				delayed = NULL;
				nf_reinject(skb, info, NF_ACCEPT);
				delay_this--;
				return 0;
			}
			else if (delay_this > 0){
				delay_this--;
			}
		}
		nf_reinject(skb, info, NF_ACCEPT);
	}
	else{
		nf_reinject(skb, info, NF_ACCEPT);
	}
	return 0;
}

/* gets called as art of sockopts */
int user_control(struct sock *sk, int cmd, void *user, unsigned int len){
	printk("Communication from user space\n");
	if (!capable(CAP_NET_ADMIN)){
		return -EPERM;
	}
	freq = ((struct freq_struct *)user)->freq;
	reorder_freq = ((struct freq_struct *)user)->reorder_freq;
	reorder_delay = ((struct freq_struct *)user)->reorder_delay;
	reorder_block = ((struct freq_struct *)user)->reorder_block;

	/* sanity check - its a simple prototype */
	while (reorder_freq <= reorder_delay + reorder_block){
		reorder_freq += 3;
	}

	drop_this = 1; /* reset counter */
	reorder_this = 1;
	delay_this = -1;
	block_size = 1;
	printk("Dropping frequency set to: %d\n", freq);
	printk("Reordering frequency set to: %d\n", reorder_freq);
	printk("Reordering delay set to: %d\n", reorder_delay);
	printk("Reordering block size set to: %d\n", reorder_block);
	return 0;
}

static struct nf_hook_ops fault_ops = {
	{ NULL, NULL },
	faulty,
	PF_INET,
	NF_IP_FORWARD,
	NF_IP_PRI_FILTER
};

/* for receiving commands from user space */
static struct nf_sockopt_ops fault_user = {
	{ NULL, NULL },
	PF_INET,
	IPT_BASE_CTL+2, /* 66 */
	IPT_BASE_CTL+3, /* 67 */
	user_control, /* function to be called */
	IPT_BASE_CTL+2, /* 66 */
	IPT_BASE_CTL+3, /* 67 */
	user_control, /* function to be called */
	0, /* these 2 used internally */
	NULL 
};

static int __init init(void){
	int queue, status, sockopt;
	printk(KERN_ERR "Starting Faulty Router - tws9@aber.ac.uk\n");

	status = nf_register_hook(&fault_ops);
	if (status < 0){
		printk(KERN_ERR "FAULTY: failed to register hook handler\n");
	}
	printk(KERN_ERR "registered hook!!\n");

	queue = nf_register_queue_handler(PF_INET, ip_fault_queue, NULL);
	if (queue < 0){
		printk(KERN_ERR "FAULTY: failed to register queue handler\n");
	}
	printk(KERN_ERR "registered queue!!\n");
	
	sockopt = nf_register_sockopt(&fault_user);
	if (sockopt < 0){
		printk(KERN_ERR "Unable to register sockopt in faulty router\n");
	}
	printk(KERN_ERR "registered sockopt: %d %d %d\n", status, sockopt, queue);
	
	/* return status of called functions */
	return status & sockopt & queue;
}

static void __exit fini(void){
	nf_unregister_queue_handle(PF_INET);
	nf_unregister_hook(&fault_ops);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
