#!/usr/bin/perl -w

# Test script to determine the extent of pack dropping, shown between two
#  tcpdump output (tcpdump -nvvv)
#       numerical output should be used for tcpdump, so that host and ports
#              numbers appear in number format

# Author:
#	Thomas William Salmon - tws9@aber.ac.uk tom@slashtom.org

# USAGE:
#  chkdrop.pl
#       [-range min,max,min,max]
#		[-icmp [min,max,min,max,....]]
#		[-igmp [min,max,min,max,....]]
#		[-tcp [min,max,min,max,....]]
#		[-udp [min,max,min,max,....]]
#		-s source_ip_address
#		-d destination_ip_address
#		sender_packet_headers.txt
#		receiver_packet_headers.txt

use strict;

sub usage{
	print " USAGE:\n";
	print "  chkdrop.pl\n";
	print "     [-range min,max,min,max]\n";
	print "		[-icmp [min,max,min,max,....]]\n";
	print "		[-igmp [min,max,min,max,....]]\n";
	print "		[-tcp [min,max,min,max,....]]\n";
	print "		[-udp [min,max,min,max,....]]\n";
	print "		-s source_ip_address\n";
	print "		-d destination_ip_address\n";
	print "		sender_packet_headers.txt\n";
	print "		receiver_packet_headers.txt\n";
	exit;
}

if ($#ARGV < 5){
	usage();
}

my $srcIP;
my $dstIP;
my $srcHeaders;
my $dstHeaders;

my $totalPacketsSent = 0;
my $totalPacketsReceived = 0;
my $minmax;

my %icmp = (
	info	=>	0,
	minmax	=>	0,
);
my %igmp = (
	info	=>	0,
	minmax	=>	0,
);
my %tcp = (
	info	=>	0,
	minmax	=>	0,
);
my %udp = (
	info	=>	0,
	minmax	=>	0,
);

# read in arguments
for (my $i=0; $i <= $#ARGV-2; $i++){
	if ($ARGV[$i] =~ m/-s/){ # tis the source ip
		$srcIP = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-d/){ # tis the destination ip
		$dstIP = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-range/){
		$minmax = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-icmp/){
		$icmp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			#$icmp{minmax} = [ split(/,/, $ARGV[$i+1]) ];
			$icmp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-igmp/){
		$igmp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$igmp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-tcp/){
		$tcp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$tcp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-udp/){
		$udp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$udp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	else{
		usage();
	}
}
#last 2 args are the headers
$srcHeaders = $ARGV[$#ARGV-1];
$dstHeaders = $ARGV[$#ARGV];


#open files containing headers
open(SRC, $srcHeaders);
open(DST, $dstHeaders);
my @src_contents = <SRC>;
my @dst_contents = <DST>;
close(SRC);
close(DST);

# read in the data from the files - only record records containing src and
#dst IP addresses (remove the noise from the rest of the network)
my @src_data = grep /$srcIP.* > $dstIP/, @src_contents;
my @dst_data = grep /$srcIP.* > $dstIP/, @dst_contents;
push @src_data, grep /$dstIP.* > $srcIP/, @src_contents;
push @dst_data, grep /$dstIP.* > $srcIP/, @dst_contents;

# strip away the time stamp in each, as that differs on the two machines
s/^[0-9]+:[0-9]+:[0-9]+\.[0-9]+ //g for @src_data;
s/^[0-9]+:[0-9]+:[0-9]+\.[0-9]+ //g for @dst_data;
# remove the ttl, as this WILL be different
s/ttl [0-9]+,//g for @src_data;
s/ttl [0-9]+,//g for @dst_data;

# strip away the MAC addresses, as these will be difference for each,
#as router will change MAC address
s/^.*:.*:.*:.*:.*:.* .*:.*:.*:.*:.*:.. //g for @src_data;
s/^.*:.*:.*:.*:.*:.* .*:.*:.*:.*:.*:.. //g for @dst_data;



# displays information by the protocol, and range
#PARAMS:
#	string - protocol (blank for all)
#	string - range of lengths: min,max,min,max....
sub proto{
	my $proto = $_[0];
	my $range = $_[1];
	my @sent = grep /$proto/, @src_data;
	my @received = grep /$proto/, @dst_data;
	#$udp{totalSent} = scalar(@sent);
	#$udp{totalReceived} = scalar(@received);
	if ($proto =~ m/win/){
		# change name to TCP 
		print("tcp:\n");
	}
	else{
		print("$proto:\n");
	}
	print("\tPackets Sent:\t".scalar(@sent)."\n");
	print("\tPackets Received:\t".scalar(@received)."\n");
	if (scalar(@sent) != 0){
		print("\t\tSuccess:\t" . (scalar(@received)/scalar(@sent))*100 ."%\n");
	}
	else{
		print("\t\tSuccess: 0%\n");
	}

	if (defined $range){
		my @minmax = split(/,/, $range);
		for (my $i=0; $i+1 < scalar(@minmax); $i=$i+2){
			print("\t\tPacket Range: $minmax[$i] to $minmax[$i+1]\n");
			my $scount=0;
			for (@sent){
				/(.*)len ([0-9]+)(.*)/;
				if ($2 > $minmax[$i] && $2 < $minmax[$i+1]){
					$scount++;
				}
			}
			print("\t\t\tSent:\t$scount\n");
		
			my $dcount=0;
			for (@received){
				/(.*)len ([0-9]+)(.*)/;
				if ($2 > $minmax[$i] && $2 < $minmax[$i+1]){
					$dcount++;
				}
			}
			print("\t\t\tReceived:\t$dcount\n");
			if ($scount == 0){
				print("\t\t\tSuccess: 0%\n");
			}
			else{
				print("\t\t\tSuccess:\t".($dcount/$scount)*100 ."%\n");
			}
		}
	}
}
	

# output section

print("Total Packets");
proto("",$minmax);

if ($icmp{info} == 1) { proto("icmp", $icmp{minmax}); }
if ($igmp{info} == 1) { proto("igmp", $igmp{minmax}); }
if ($udp{info} == 1) { proto("udp", $udp{minmax}); }
if ($tcp{info} == 1) { proto("win", $tcp{minmax}); } # tcp
