#!/usr/bin/perl -w

# Test script to determine the extent of pack dropping, shown between two
#  tcpdump output (tcpdump -nvvv)
#       numerical output should be used for tcpdump, so that host and ports
#              numbers appear in number format

# Author:
#	Thomas William Salmon - tws9@aber.ac.uk tom@slashtom.org

# USAGE:
#  chkdrop.pl
#       [-range min,max,min,max]
#		[-icmp [min,max,min,max,....]]
#		[-igmp [min,max,min,max,....]]
#		[-tcp [min,max,min,max,....]]
#		[-udp [min,max,min,max,....]]
#		-s source_ip_address
#		-d destination_ip_address
#		sender_packet_headers.txt
#		receiver_packet_headers.txt

use strict;

sub usage{
	print " USAGE:\n";
	print "  chkdrop.pl\n";
	print "     [-range min,max,min,max]\n";
	print "		[-icmp [min,max,min,max,....]]\n";
	print "		[-igmp [min,max,min,max,....]]\n";
	print "		[-tcp [min,max,min,max,....]]\n";
	print "		[-udp [min,max,min,max,....]]\n";
	print "		-s source_ip_address\n";
	print "		-d destination_ip_address\n";
	print "		sender_packet_headers.txt\n";
	print "		receiver_packet_headers.txt\n";
	exit;
}

if ($#ARGV < 5){
	usage();
}

my $srcIP;
my $dstIP;
my $srcHeaders;
my $dstHeaders;

my $minmax;

my %icmp = (
	info	=>	0,
	minmax	=>	0,
);
my %igmp = (
	info	=>	0,
	minmax	=>	0,
);
my %tcp = (
	info	=>	0,
	minmax	=>	0,
);
my %udp = (
	info	=>	0,
	minmax	=>	0,
);

# read in arguments
for (my $i=0; $i <= $#ARGV-2; $i++){
	if ($ARGV[$i] =~ m/-s/){ # tis the source ip
		$srcIP = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-d/){ # tis the destination ip
		$dstIP = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-range/){
		$minmax = $ARGV[$i+1];
		$i++;
	}
	elsif ($ARGV[$i] =~ m/-icmp/){
		$icmp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$icmp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-igmp/){
		$igmp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$igmp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-tcp/){
		$tcp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$tcp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	elsif ($ARGV[$i] =~ m/-udp/){
		$udp{'info'} = 1;
		if (!($ARGV[$i+1] =~ m/^-.*/)){
			# we have min max ranges
			$udp{minmax} = $ARGV[$i+1];
			$i++;
		}
	}
	else{
		usage();
	}
}
#last 2 args are the headers
$srcHeaders = $ARGV[$#ARGV-1];
$dstHeaders = $ARGV[$#ARGV];


#open files containing headers
open(SRC, $srcHeaders);
open(DST, $dstHeaders);
my @src_contents = <SRC>;
my @dst_contents = <DST>;
close(SRC);
close(DST);

# read in the data from the files - only record records containing src and
#dst IP addresses (remove the noise from the rest of the network)
my @src_data1 = grep /$dstIP/, @src_contents;
my @dst_data1 = grep /$srcIP/, @dst_contents;
my @src_data = grep /$dstIP/, @src_data1;
my @dst_data = grep /$dstIP/, @dst_data1;

# strip away the time stamp in each, as that differs on the two machines
s/^[0-9]+:[0-9]+:[0-9]+\.[0-9]+ //g for @src_data;
s/^[0-9]+:[0-9]+:[0-9]+\.[0-9]+ //g for @dst_data;
# remove the ttl, as this WILL be different
s/ttl [0-9]+,//g for @src_data;
s/ttl [0-9]+,//g for @dst_data;

# strip away the MAC addresses, as these will be difference for each,
#as router will change MAC address
s/^.*:.*:.*:.*:.*:.* .*:.*:.*:.*:.*:.. //g for @src_data;
s/^.*:.*:.*:.*:.*:.* .*:.*:.*:.*:.*:.. //g for @dst_data;

my @dst_orig = @dst_data; # take a backup of dst data


my @reordered_packets;

# records the delay experienced
#	PARAMS:	String line, int delay, int burst, int interval
sub record{
	# print("RECORD:\t$_[1] : $_[2] : $_[3] : $_[0]\n");
	push @reordered_packets, "$_[1] : $_[2] : $_[3] : $_[0]";
}


my $s = 0;;
my $d = 0;
my $orig_d = 0;
my $last = 0; # index of last packet taken out of order
my $b = 0;;
my $id;
my $delay;
my $tmp_d;
my $tmp_s;

my $max_d = scalar(@dst_data);
my $max_s = scalar(@src_data);
my $max_delay = 200; #should not exceed this

my $last_burst = 0;

# synchronise data first
#while($d < $max_d-1 && $src_data[$s] ne $dst_data[$d]){
#	$d++;
#}

for (; $s<$max_s && $d<$max_d; $s++){
	# check for packet drop
	if (scalar (grep /\Q$src_data[$s]/, @dst_data) == 0){
		next;
	}
	$delay = 0;
	$orig_d = $d;
#print("$s\t$d\n");
#print(" $src_data[$s] $dst_data[$d] \n");
	while($delay < $max_delay && $d < $max_d && $src_data[$s] ne $dst_data[$d]){
		$d++;
		$delay++;
	}
	if ($delay > 0 && $delay < $max_delay && $delay < $max_d){
#print(" $src_data[$s] $dst_data[$d] Delay:\t$delay\n\n");
		if ($d <= $last + $b){
			#$last++;
			$d = $orig_d + 1;
			next;
		}

		$b = 1; # set burst to one
		$tmp_d = $d+1;
		$tmp_s = $s+1;
		# skip over packet dropped entries
		while (scalar (grep /\Q$src_data[$tmp_s]/, @dst_data) == 0){
			$tmp_s++;
		}
		while ($tmp_s < $max_s && $tmp_d < $max_d &&
				$src_data[$tmp_s] eq $dst_data[$tmp_d]){
			$b++; # increment burst
			$tmp_d++;
			$tmp_s++;

			# skip of packet dropped stuff
			while (scalar (grep /\Q$src_data[$tmp_s]/, @dst_data) == 0){
				$tmp_s++;
			}
		}

		# record the burst
		if ($d == $max_d){ $d = $orig_d + 1; next; }
#		print("$dst_data[$d], $delay, $b, ($d - $last_burst)\n");
		if ($last_burst == 0){ # special case
			record($dst_data[$d], $delay, $b, ($d - $last_burst - $delay + 1));
		}
		else{
			record($dst_data[$d], $delay, $b, ($d - $last_burst));
		}
		$last_burst = $d + $b - 1;
		$last = $d;

		$d = $orig_d + 1;
		next;
	}
	$d = $orig_d + 1;
}

#print("$s\t$d\n");
#print(" $src_data[$s] $dst_data[$d] \n");

# gets stats of @_
# returns average_interval, average_burst, average_delay
sub get_stats{
	my $total_entries = scalar(@_);
	my $total_delay = 0;
	my $total_burst = 0;
	my $total_interval = 0;
	my $i;
	my $j;
	my $k;
	my $l;
	my $loop_count = 0;
	for (@_){
		($i, $j, $k, $l) = split(/:/, $_, 4);
		$loop_count++;
		$total_delay = $total_delay + $i;
		$total_burst = $total_burst + $j;
		if ($loop_count == 1){
			next; # discount the first interval reading
		}
		$total_interval = $total_interval + $k;
	}

	if ($total_entries > 0){
		return ($total_interval/$total_entries,
				$total_burst/$total_entries,
				$total_delay/$total_entries);
	}
	else{
		return (0, 0, 0);
	}
}



sub proto{
    my $proto = $_[0];
    my $range = $_[1];
    my @reorder = grep /$proto/, @reordered_packets;
    if ($proto =~ m/win/){
        # change name to TCP
        print("tcp:\n");
    }
    else{
        print("$proto:\n");
    }

    #form some statistics as regards reorder interval, burst and delay averages

	my $aInt;
	my $aBurst;
	my $aDelay;
	($aInt, $aBurst, $aDelay) = get_stats(@reorder);
    print("\tAverage reorder interval:\t".$aInt."\n");
    print("\tAverage burst:\t".$aBurst."\n");
    print("\tAverage delay:\t".$aDelay."\n");

    if (defined $range){
        my @minmax = split(/,/, $range);
        for (my $i=0; $i+1 < scalar(@minmax); $i=$i+2){
            my @reorder_size;
            print("\t\tPacket Range: $minmax[$i] to $minmax[$i+1]\n");
            my $scount=0;
   
            for (@reorder){
                /(.*)len ([0-9]+)(.*)/;
                if ($2 > $minmax[$i] && $2 < $minmax[$i+1]){
                    push @reorder_size, $_; # add this entry to array
                }
            }
			($aInt, $aBurst, $aDelay) = get_stats(@reorder_size);
		    print("\t\t\tAverage reorder interval:\t".$aInt."\n");
		    print("\t\t\tAverage burst:\t".$aBurst."\n");
		    print("\t\t\tAverage delay:\t".$aDelay."\n");
   
            undef &reorder_size; # clear old array
        }
    }
}
# output section

print("Total Packets");
proto("",$minmax);

if ($icmp{info} == 1) { proto("icmp", $icmp{minmax}); }
if ($igmp{info} == 1) { proto("igmp", $igmp{minmax}); }
if ($udp{info} == 1) { proto("udp", $udp{minmax}); }
if ($tcp{info} == 1) { proto("win", $tcp{minmax}); } # tcp


